
(*********************************************************************)
(* Title: Generalized summation by parts algorithm                   *)
(* Program: psum.m                                                   *)
(* Created: 4 July 2014                                              *)
(* Version: 1.2 (19 March 2015, at 09.02)                            *)
(* Author: Vsevolod E. Adler                                         *)
(* Language: Mathematica 5.0                                         *)
(* Summary: psum[m,a,b] solves the difference equation               *)
(*         T[y,m]-a*y=b,                                             *)
(*         where T is the shift operator u[j] -> u[j+1]              *)
(* Input:  m is a natural number,                                    *)
(*         a,b are expressions depending on finite number of u[j]    *)
(* Output: {y,obstacle}                                              *)
(*         if obstacle=0 then y is the general solution              *)
(* Refs: V.E. Adler. Integrability test for evolutionary             *)
(*         lattice equations of higher order.                        *)
(*         http://arxiv.org/abs/1408.5726                            *)
(* Limitations: The algorithm relies upon the functions              *)
(*         cf[f] (canonical form) and X[f,k] (extraction).           *)
(*         Their default definitions are suitable for equations      *)
(*         with rational coefficients. In a more general situation,  *)
(*         this may require redefining.                              *)
(*********************************************************************)

BeginPackage["psum`"]

u::usage="u[j] are the dynamical variables"
vars::usage="vars[expr] returns the list of variables u[j] in expression \
(without simplification)"
ords::usage="ords[expr] returns the minimal and maximal j such that expr \
depends on u[j]"
T::usage="T is the shift operator u[j]->u[j+1]"
psum::usage="psum[m,a,b] solves the difference equation T[y,m]-a*y=b. The \
output is {y, obstacle}"
const::usage="const is the `integration' constant"
cf::usage="cf[expr] is the canonical form of expr"
X::usage="X[expr,k] extracts terms with u[k]"

Unprotect[u,vars,ords,T,psum,const]

Begin["`Private`"]

T[f_,k_]:=f/.u[j_]\[RuleDelayed]u[j+k]

cf[f_]:=Together[f]
X[f_,k_]:=Integrate[cf[D[f,u[k]]],u[k]]

vars[f_]:=Union[Cases[f,_u,{0,Infinity}]]

ords[f_]:=Block[{v=vars[f],lo,hi,j},
    hi=0;
    Do[If[UnsameQ[cf[D[f,v[[j]]]],0],hi=j;Break[]],{j,Length[v],1,-1}];
    Do[If[UnsameQ[cf[D[f,v[[j]]]],0],lo=j;Break[]],{j,1,hi}];
    If[hi\[Equal]0,{Infinity,-Infinity},{v[[lo,1]],v[[hi,1]]}]
    ]

(***** main function *****)

psum[m_Integer,aa_,bb_]/;m>0:=Module[
    {a=cf[aa],b=cf[bb],A,B,p1,p2,q1,q2,r,y},
    
    q2=ords[a];q1=q2[[1]];q2=q2[[2]];
    p2=ords[b];p1=p2[[1]];p2=p2[[2]];
    
    Catch[
      If[a===0,Throw[{T[b,-m],0}]];
      
      (***** Case a=const *****)
      If[q1\[Equal]Infinity,
        If[p2<p1+m,
          y=If[a===1,const,b/(1-a)];
          Throw[cf[{y,T[y,m]-a*y-b}]]
          ];
        
        (* B=cf[D[b,u[p2]]]; *)
        r=ords[D[b,u[p2]]][[1]];
        If[r<p1+m,Throw[{0,D[b,u[p2],u[r]]}]];
        B=X[b,p2];
        Throw[psum[m,a,b-B+a*T[B,-m]]+{T[B,-m],0}]
        ];
      
      (***** Reduction of a *****)
      (* A=cf[D[a,u[q1]]/a]; *)
      r=ords[D[a,u[q1]]/a][[2]];
      If[And[q1\[LessEqual]q2-m,r\[LessEqual]q2-m],
        A=Exp[X[Log[a],q1]];
        Throw[psum[m,a*T[A,m]/A,b*T[A,m]]/A]
        ];
      
      (***** Reduction of b *****)
      If[p2>q2,
        B=X[b,p2];
        Throw[psum[m,a,b-B+a*T[B,-m]]+{T[B,-m],0}]
        ];
      
      If[p1<q1,
        B=cf[D[b,u[p1]]/a];
        r=ords[B][[2]];
        If[r>q2-m,Throw[{0,D[B,u[r]]}]];
        B=X[b/a,p1];
        Throw[psum[m,a,b+T[B,m]-a*B]-{B,0}]
        ];
      
      (***** Solving of linear system *****)
      y=If[q1\[LessEqual]q2-m,
          -(D[b,u[q1]]*D[a,u[r]]-D[b,u[q1],u[r]]*a)/(D[a,u[q1]]*D[a,u[r]]-
                D[a,u[q1],u[r]]*a),
          b/(1-a)
          ];
      Throw[cf[{y,T[y,m]-a*y-b}]]
      ]
    ]

End[]

Protect[u,vars,ords,T,psum,const]

EndPackage[]